/**BGLayers library
@author Billy D. Spelchan
@version 0.1
@license
Copyright (c) 2010 Billy D. Spelchan.
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 THE SOFTWARE.
*/

// Create the namespace for this library
var BGLayers = new Object();

/**
 * General purpose function used to copy the properties from a parent or super class
 * into it's child. This is used to add inheritance to JavaScript in a way that is
 * compatible with most browsers.
 *
 * @param child The child class that is going to inherit properties.
 * @param supr The Super class from which properties will be inherited.
*/
BGLayers.inheritProperties = function(child, supr)
{
	for (var property in supr) {  
		if (typeof child[property] == "undefined")  
			child[property] = supr[property];  
	}  
}

// --------------------------------------------------------------------------

/**
* @class Generic point class for indicating the position of a point on a 2D plane.
*/
BGLayers.Point = function(x,y)
{
	this.x = x;
	this.y = y;
}

// --------------------------------------------------------------------------

/**
* @class Generic dimension class for describing the width and hight of something.
*/
BGLayers.Dimension = function(w,h)
{
	this.width = w;
	this.height = h
}

// --------------------------------------------------------------------------

/**
* @class Rectangle. Passing no parameters creates an empty rectangle. Passing
* a rectangle clones the rectangle. Passing x,y,width,height creates the 
* specified rectangle.
*/
BGLayers.Rectangle = function(x, y, w, h)
{
	if (arguments.length == 1) { // clone rectangle
		this.x = x.x;
		this.y = x.y;
		this.width = x.width;
		this.height = x.height;
	} else if (arguments.length == 4) { // parameterized rectangle
		this.x = x;
		this.y = y;
		this.width = w;
		this.height = h;
	} else { // default to empty rectangle
		this.x = 0;
		this.y = 0;
		this.width = 0;
		this.height = 0;
	}
}

/**
* Utility method to change the bounds of a rectangle.
*/
BGLayers.Rectangle.prototype.bounds = function(x,y,w,h)
{
	this.x = x;
	this.y = y;
	this.width = w;
	this.height = h;
}

/**
* Checks to see if the indicated point is inside the rectangle returning true if it is.
*/
BGLayers.Rectangle.prototype.contains = function(x,y) {
	var deltaX = x - this.x;
	var deltaY = y - this.y;
	if ((deltaX < 0) || (deltaY < 0) || (deltaX > this.width) || (deltaY > this.height))
		return false;
	else
		return true;
}

/**
* Creates a rectangle that is the Intersection of two rectangles.
*/
BGLayers.Rectangle.prototype.getIntersection = function(rect)
{
	var x = Math.max(this.x, rect.x);
	var x2 = Math.min(this.x+this.width, rect.x + rect.width);
	var y = Math.max(this.y, rect.y);
	var y2 = Math.min(this.y+this.height, rect.y + rect.height);
	return new BGLayers.Rectangle(x, y, x2-x, y2-y);
}

/**
* Checks to see if there is any overlapping between this rectangle and the passed rectangle
* returning true if there is some overlap.
*/
BGLayers.Rectangle.prototype.intersects = function(rect) 
{
	if ( (this.x < (rect.x + rect.width)) && 
			( (this.x + this.width) > rect.x) &&
			(this.y < (rect.y + rect.height)) && 
			( (this.y + this.height) > rect.y) )
		return true;
	else
		return false;
}

/**
 * Align the rectangle to pixel bounds (parameters are all integers).
*/
BGLayers.Rectangle.prototype.makePixelAligned = function()
{
	var x = Math.floor(this.x);
	var y = Math.floor(this.y);
	var w = Math.ceil(this.width + (this.x - x));
	var h = Math.ceil(this.height + (this.y - y));
	this.bounds(x,y,w,h);
}

// --------------------------------------------------------------------------

/**
 * @class The most basic of layers used to draw a solid rectangle. This has built in
 * child handling so can be used as a container as well.
 * 
 * @id an object used to identify this particular layer.
 * @w The logical width of the layer
 * @h The logical height of the layer
 */
BGLayers.Layer = function(id, w, h)
{
	this.id = id;
	this.logicalSize = new BGLayers.Dimension(w, h);
	this.logicalPosition = new BGLayers.Rectangle(0,0,w,h);
	this.realPosition = new BGLayers.Rectangle();
	this.realPositionKnown = false;
	this.solid = true;
	this.backgroundColor = 'rgb(0,0,0)';
	this.children = new Array();
	this.visible = true;
	this.dirtyList = new Array();
	this.dirtyList.push(this.logicalPosition);
}

/**
* Adds a child to the list of children that this layer has.
*
* @param child Child to add to the display list.
*/
BGLayers.Layer.prototype.addChild = function(child, logPos) 
{ 
	child.setParent(this, logPos);
	this.children.push(child); 
}

/**
* Adds a rectangle to the top parent's dirty rectangle list. The rectangle should be in 
* real-world coordinates. Pass null for the rectangle to force the entire layer to become dirty.
*/
BGLayers.Layer.prototype.addDirty = function(rect)
{
	var dirty;
	if (rect == null) // force full dirty layer
		dirty = new BGLayers.Rectangle(this.findRealPosition());
	else
		dirty = new BGLayers.Rectangle(rect);
	dirty.makePixelAligned();
	if (this.parent == null) {
		// todo loop through all rectangles to eliminate overlap.
		this.dirtyList.push(dirty);
	} else 
		this.parent.addDirty(dirty);
}

/**
* Changes the logical bounds of the layer. Passing a null value forces the layer to
* recalculate it's real position.
*/
BGLayers.Layer.prototype.adjustPosition = function(logPos)
{
	if (this.parent != null)
		this.parent.addDirty(this.findRealPosition);
	if (logPos != null) {
		this.logicalPosition.x = logPos.x;
		this.logicalPosition.y = logPos.y;
		this.logicalPosition.width = logPos.width;
		this.logicalPosition.height = logPos.height;
	}
	this.realPositionKnown = false;
	for (var cntr = 0; cntr < this.children.length; ++cntr)
		this.children[cntr].adjustPosition(null);
}

/**
* Converts the provided rectangle from layer coordinates to real canvas coordinates.
*/
BGLayers.Layer.prototype.convertLogicalToReal = function(rect)
{
	var realRect = new Rectangle();
	this.findRealPosition();
	var scaleX = this.realPosition.width / this.logicalSize.width;
	var scaleY = this.realPosition.height / this.logicalSize.height;
	realRect.x = this.realPosition.x + rect.x * scaleX;
	realRect.y = this.realPosition.y + rect.y * scaleY;
	realRect.width = this.realPosition.width + rect.width * scaleX;
	realRect.height = this.realPosition.height + rect.height * scaleY;
	return realRect;
}

/**
* Draws itself (not including children).
* @protected This should only be called internally
* @param ctx Canvas context
* @param bounds Real boundaries of the area to be drawn
*/
BGLayers.Layer.prototype.drawSelf = function(ctx, bounds)
{
	if (this.findRealPosition().intersects(bounds) == false)
		return;
	if (this.solid) {
		ctx.save();
		this.findRealPosition();
		var rect = this.realPosition.getIntersection(bounds);
		var oldFill = ctx.fillStyle;
		ctx.fillStyle = this.backgroundColor;
		ctx.fillRect(rect.x, rect.y, rect.width, rect.height);
		ctx.fillStyle = oldFill;
		ctx.restore();
	}
}

/**
* Translates the logical bounds into real canvas coordinates
*/
BGLayers.Layer.prototype.findRealPosition = function()
{
	if (this.realPositionKnown)
		return this.realPosition;
	var parentPos = this.logicalPosition;
	var w = this.logicalSize.width;
	var h = this.logicalSize.height;
	if (this.parent != null) {
		parentPos = this.parent.findRealPosition();
		w = this.parent.logicalSize.width;
		h = this.parent.logicalSize.height;
	}
	var scaleX = parentPos.width / w;
	var scaleY = parentPos.height / h;
	this.realPosition.x = parentPos.x + this.logicalPosition.x * scaleX;
	this.realPosition.y = parentPos.y + this.logicalPosition.y * scaleY;
	this.realPosition.width = this.logicalPosition.width * scaleX;
	this.realPosition.height = this.logicalPosition.height * scaleY;
	this.realPositionKnown = true;
	return this.realPosition;
}


BGLayers.Layer.prototype.mouseDown = function(x, y)
{
	var dirty = false;
	for (var cntr = 0; cntr < this.children.length; ++cntr)
		dirty |= this.children[cntr].mouseDown(x,y);
	return dirty;
}

BGLayers.Layer.prototype.mouseMove = function(x, y)
{
	var dirty = false;
	for (var cntr = 0; cntr < this.children.length; ++cntr)
		dirty |= this.children[cntr].mouseMove(x,y);
	return dirty;
}

BGLayers.Layer.prototype.mouseUp = function(x, y)
{
	var dirty = false;
	for (var cntr = 0; cntr < this.children.length; ++cntr)
		dirty |= this.children[cntr].mouseUp(x,y);
	return dirty;
}

/**
* Changes the logical position of this layer.
*/
BGLayers.Layer.prototype.moveTo = function(x,y)
{
	this.addDirty(new BGLayers.Rectangle(this.findRealPosition()));
	this.logicalPosition.x = x;
	this.logicalPosition.y = y;
	this.realPositionKnown = false;
	for (var cntr = 0; cntr < this.children.length; ++cntr)
		this.children[cntr].adjustPosition(null);	
	this.addDirty(new BGLayers.Rectangle(this.findRealPosition()));
}

/**
* Removes the indicated child from the display list.
* 
* @param child Child to remove from the display list.
*/	
BGLayers.Layer.prototype.removeChild = function(child)
{
	var childIndex = -1;
	var cntr;
	
	addDirty(child.realPosition);
	for (cntr = 0; cntr < this.children.length; ++cntr)
		if (this.children[cntr] == child)
			childIndex = cntr;
	if (childIndex >= 0)
		this.children.splice(childIndex, 1);
	child.setParent(null, child.logicalPosition);
}

/**
* Draws the layer and it's children onto the canvas.
*
* @param ctx Canvas context
* @param bounds Real boundaries of the area to be drawn
*/
BGLayers.Layer.prototype.render = function(ctx, bounds)
{
	if (!this.visible)
		return;
	this.drawSelf(ctx, bounds);
	for (var cntr = 0; cntr < this.children.length; ++cntr)
		this.children[cntr].render(ctx, bounds);
}

/**
* Draws only the dirty areas of the layer.
*/
BGLayers.Layer.prototype.renderDirty = function(ctx)
{
	while (this.dirtyList.length > 0) {
		var bounds = this.dirtyList.pop();
		this.render(ctx, bounds);
	}
}

/**
* Sets the parent of this layer. The parent is what the logical position of the  layer
* is relative to.
*/
BGLayers.Layer.prototype.setParent = function(parent, logPos)
{
	this.adjustPosition(logPos);
	if (this.parent != null)
		this.parent.removeChild(this)
	this.parent = parent;
}

// --------------------------------------------------------------------------


// --------------------------------------------------------------------------
